from .imports import *
from .hash_utils import *
##from .dynimport import get_abstract_import,import_symbols_to_parent,call_for_all_tabs
from .json_utils import (unified_json_loader,
                         find_keys,
                         get_key_values_from_path,
                         get_value_from_path,
                         find_paths_to_key,
                         create_and_read_json,
                         unified_json_loader,
                         all_try,
                         try_json_loads,
                         get_error_msg,
                         get_any_key,
                         get_any_value,
                         json_key_or_default,
                         format_json_key_values,
                         get_all_keys,
                         update_dict_value,
                         get_all_key_values,
                         get_json_data,
                         safe_json_dumps,
                         safe_json_loads,
                         read_from_json,
                         safe_load_from_json,
                         safe_load_from_file,
                         safe_json_reads,
                         safe_read_from_json,
                         safe_dump_to_file,
                         safe_dump_to_json,
                         safe_write_to_json,
                         safe_write_to_file,
                         safe_save_updated_json_data,
                         get_result_from_data,
                         flatten_json,
                         to_json_safe
                         )

from .directory_utils import *
from .list_utils import (get_highest_value_obj,
                         make_list,
                         safe_list_return,
                         get_actual_number,
                         compare_lists,
                         get_symetric_difference,
                         list_set,
                         make_list_it,
                         get_single_from_list
                         )
from .time_utils import (get_time_stamp,
                         get_sleep,
                         sleep_count_down,
                         get_date,
                         get_current_time_with_delta,
                         format_timestamp,
                         timestamp_to_milliseconds,
                         parse_timestamp,
                         get_time_now_iso,
                         is_valid_time)
from .string_utils import (eatInner,
                           eatAll,
                           eatOuter,
                           url_join,
                           capitalize
                           )
from .type_utils import (make_bool,
                         T_or_F_obj_eq,
                         is_number,
                         makeInt,
                         str_lower,
                         confirm_type,
                         get_all_file_types,
                         is_media_type,
                         get_bool_response,
                         get_mime_type,
                         get_media_exts,
                         if_true_get_string,
                         find_for_string,
                         is_strings_in_string,
                         get_alphabet_str,
                         get_alphabet_upper_str,
                         get_alphabet_comp_str,
                         get_numbers,
                         get_numbers_comp,
                         is_any_instance,
                         break_string,
                         MIME_TYPES,
                         get_if_None,
                         derive_media_type,
                         get_media_map
                         )
get_media_types = get_all_types = get_all_file_types
from .math_utils import (convert_to_percentage,
                         exponential,
                         get_percentage,
                         add_it,
                         divide_it,
                         exp_it,
                         return_0,
                         zero_buff_int)
from .compare_utils import (create_new_name,
                            get_last_comp_list,
                            get_closest_match_from_list,
                            get_first_match,
                            get_all_match,
                            best_match)
from .thread_utils import ThreadManager
from .history_utils import HistoryManager

from .parse_utils import (num_tokens_from_string,
                          chunk_source_code,
                          chunk_any_to_tokens,
                          detect_language_from_text,
                          chunk_by_language_context,
                          search_code,
                          get_within_quotes)

from .log_utils import get_caller_info,get_logFile,print_or_log,get_json_call_response,initialize_call_log
from .error_utils import try_func
from .class_utils import *
from .ssh_utils import *
from .env_utils import *

from .file_utils import *
from .string_utils import *
from .import_utils import *
from .path_utils import *
from .read_write_utils import (read_from_file,
                               write_to_file,
                               make_dirs,
                               make_dirs,
                               copy_files,
                               make_path,
                               run_cmd
                               )



from .imports import *
def full_hash(path):
    """Only called when quick hashes match"""
    sha256 = hashlib.sha256()
    with open(path, 'rb') as f:
        while chunk := f.read(8192):
            sha256.update(chunk)
    return sha256.hexdigest()
def quick_hash(path, bytes_to_read=16384):
    """Fast fingerprint to avoid full hash on obvious non-dups"""
    with open(path, 'rb') as f:
        return hashlib.sha256(f.read(bytes_to_read)).hexdigest()
def generate_data_hash(insertName,value):
    # Combine values to create a unique reference
    data_string = f"{insertName}_{value}"
    return hashlib.md5(data_string.encode()).hexdigest()
def get_size_to_paths(directory,exts):
    exts=make_list(exts or [])
    size_to_paths = defaultdict(list)
    dirs,file_paths = get_files_and_dirs(directory,allowed_exts=exts)
    for file_path in file_paths:
        if os.path.isfile(file_path):
            size_to_paths[os.path.getsize(file_path)].append(file_path)
    return size_to_paths
def dedupe_media(directory,exts):
    size_to_paths = get_size_to_paths(directory,exts)
    hash_to_paths = defaultdict(list)
    for size_group in size_to_paths.values():
        if len(size_group) < 2: continue
        quick_to_paths = defaultdict(list)
        for p in size_group:
            if os.path.isfile(p):
                quick_to_paths[quick_hash(p)].append(p)
        for quick_matches in quick_to_paths.values():
            if len(quick_matches) < 2: continue
            full_h = full_hash(quick_matches[0])
            hash_to_paths[full_h].extend(quick_matches)
    return hash_to_paths
def dedupe_hash_to_paths(directory,exts):
    hash_to_paths_js = {}
    hash_to_paths = dedupe_media(directory,exts)
    hash_to_paths_js["hash_to_paths"] = hash_to_paths
    hash_to_paths_js["dedups"] = {h: paths for h, paths in hash_to_paths.items() if len(paths) >= 2}
    hash_to_paths_js["singles"] = {h: paths for h, paths in hash_to_paths.items() if len(paths) < 2}
    return hash_to_paths_js
def dedupe_and_save_hash_to_paths(directory,exts,basename=None):
    basename = basename or 'hash_to_paths.json'
    hash_to_paths_js = dedupe_hash_to_paths(directory,exts)
    file_path = os.path.join(directory,basename)
    safe_dump_to_json(data=hash_to_paths_js,file_path=file_path)
    return file_path



from ...imports import *
from .nullProxy import nullProxy,nullProxy_logger
from ..import_utils import inject_from_imports_map,get_imports
import importlib,importlib.util, pkgutil,sys
from pathlib import Path


def dynamic_import(target: str | Path):
    """
    Import module/package/file dynamically.
    Returns module or nullProxy if unavailable.
    """
    try:
        p = Path(target)

        # file import
        if p.exists():
            name = f"_dyn_{p.stem}_{hash(p)}"

            spec = importlib.util.spec_from_file_location(name, str(p))
            if spec is None or spec.loader is None:
                raise ImportError(target)

            mod = importlib.util.module_from_spec(spec)
            sys.modules[name] = mod
            spec.loader.exec_module(mod)

            return mod

        # module import
        return importlib.import_module(target)

    except Exception:
        return nullProxy(str(target))
def safe_attr(obj, attr):
    try:
        return getattr(obj, attr)
    except Exception:
        return nullProxy(f"{obj}.{attr}")
@lru_cache(maxsize=None)
def lazy_import_single(name: str,fallback=None):
    """
    Import module safely. If unavailable, return NullProxy.
    """

    if name in sys.modules:
        return sys.modules[name]

    try:
        module = importlib.import_module(name)
        return module
    except Exception as e:
        nullProxy_logger.warning(
            "[lazy_import] Failed to import '%s': %s",
            name,
            e,
        )
        return nullProxy(name,fallback=fallback)

def get_lazy_attr(module_name: str, *attrs,fallback=None):
    obj = lazy_import(module_name,fallback=fallback)

    for attr in attrs:
        try:
            obj = getattr(obj, attr)
        except Exception:
            return nullProxy(module_name, attrs,fallback=fallback)

    return obj
def lazy_import(name: str, *attrs,fallback=None):
    """
    Import module safely. If unavailable, return NullProxy.
    """
    if attrs:
        obj = get_lazy_attr(name, *attrs,fallback=fallback)
    else:
        obj = lazy_import_single(name,fallback=fallback)
    return obj
def lazy_import_star(files, *, self=None):
    imports_map = get_imports(files)

    return inject_from_imports_map(
        imports_map,
        into=globals(),
        self=self
    )
def lazy_import_star(targets, *, into=None, self=None):
    """
    Supports:
    - module names
    - packages
    - filesystem paths
    """

    import importlib
    import pkgutil
    from pathlib import Path

    ns = into or globals()

    for target in targets:

        # ----------------------------
        # MODULE / PACKAGE IMPORT
        # ----------------------------
        try:
            mod = importlib.import_module(target)

            if hasattr(mod, "__path__"):
                # package → import all submodules
                for _, name, _ in pkgutil.iter_modules(mod.__path__):
                    sub = importlib.import_module(f"{mod.__name__}.{name}")
                    ns[name] = sub
            else:
                # single module
                ns[target.split(".")[-1]] = mod

            continue

        except Exception:
            pass

        # ----------------------------
        # FILESYSTEM IMPORT
        # ----------------------------
        p = Path(target)

        if p.exists():
            imports_map = get_imports([str(p)])
            inject_from_imports_map(imports_map, self=self)

        else:
            nullProxy_logger.warning(f"[lazy_import] Could not import {target}")



from .imports import *

class LevelFilter(logging.Filter):
    """Filter that allows selective level enablement/disablement."""
    
    def __init__(self):
        super().__init__()
        self.enabled_levels = {
            logging.DEBUG,
            logging.INFO,
            logging.WARNING,
            logging.ERROR,
            logging.CRITICAL,
        }
    
    def filter(self, record: logging.LogRecord) -> bool:
        return record.levelno in self.enabled_levels
    
    def enable_level(self, level: int) -> None:
        self.enabled_levels.add(level)
    
    def disable_level(self, level: int) -> None:
        self.enabled_levels.discard(level)
    
    def is_enabled(self, level: int) -> bool:
        return level in self.enabled_levels


class AbstractLogManager(metaclass=SingletonMeta):
    def __init__(self):
        self.logger = logging.getLogger("AbstractLogManager")
        self.logger.setLevel(logging.DEBUG)  # Logger accepts everything
        
        self.console_handler = logging.StreamHandler()
        self.console_handler.setLevel(logging.DEBUG)  # Handler accepts everything
        
        # Filter does the actual control
        self.level_filter = LevelFilter()
        self.console_handler.addFilter(self.level_filter)
        
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        self.console_handler.setFormatter(formatter)
        
        if not self.logger.hasHandlers():
            self.logger.addHandler(self.console_handler)
    
    def set_debug(self, enabled: bool) -> None:
        """Enable or disable DEBUG level messages."""
        if enabled:
            self.level_filter.enable_level(logging.DEBUG)
            self.logger.debug("DEBUG logging enabled.")
        else:
            self.level_filter.disable_level(logging.DEBUG)
            self.logger.info("DEBUG logging disabled.")
    
    def set_info(self, enabled: bool) -> None:
        """Enable or disable INFO level messages."""
        if enabled:
            self.level_filter.enable_level(logging.INFO)
            self.logger.info("INFO logging enabled.")
        else:
            self.level_filter.disable_level(logging.INFO)
            self.logger.warning("INFO logging disabled.")
    
    def set_warning(self, enabled: bool) -> None:
        """Enable or disable WARNING level messages."""
        if enabled:
            self.level_filter.enable_level(logging.WARNING)
            self.logger.warning("WARNING logging enabled.")
        else:
            self.level_filter.disable_level(logging.WARNING)
            self.logger.error("WARNING logging disabled.")
    
    def set_error(self, enabled: bool) -> None:
        """Enable or disable ERROR level messages."""
        if enabled:
            self.level_filter.enable_level(logging.ERROR)
        else:
            self.level_filter.disable_level(logging.ERROR)
    
    def is_level_enabled(self, level: int) -> bool:
        """Check if a level is currently enabled."""
        return self.level_filter.is_enabled(level)
    
    def get_logger(self) -> logging.Logger:
        """Return the configured logger instance."""
        return self.logger



from time import time
import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="abstract_utilities",
    version='0.2.2.772',
    author="putkoff",
    author_email="partners@abstractendeavors.com",
    description="Utility modules for data comparison, JSON handling, string manipulation, math operations, and general automation tasks.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AbstractEndeavors/abstract_utilities",

    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],

    python_requires=">=3.10",

    install_requires=[
        "lxml",
        "ezodf",
        "tiktoken",
        "geopandas",
        "abstract_windows",
        "PyPDF2",
        "pdfplumber",
        "pytesseract",
        "flask",
        "pdf2image",
        "abstract_paths",
        "pathlib>=1.0.1",
        "abstract_security>=0.0.1",
        "yt_dlp>=2023.10.13",
        "pexpect>=4.8.0",
    ],

    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
)



from .imports import *
from .module_imports import read_from_file, eatAll,make_list
from .constants import *

# ============================================================
# Helpers
# ============================================================
def get_caller_path(i=None):
    i = i or 1
    frame = inspect.stack()[i]
    return os.path.abspath(frame.filename)



def eatElse(stringObj, chars=None):
    chars = make_list(chars or []) + list('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_')
    while stringObj:
        if stringObj and stringObj[0] not in chars:
            stringObj = stringObj[1:]
            continue
        if stringObj and stringObj[-1] not in chars:
            stringObj = stringObj[:-1]
            continue
        break
    return stringObj

def clean_line(line):
    return eatAll(line, [' ', '', '\t', '\n'])

def is_line_import(line):
    return bool(line and line.startswith(import_tag) and 'from ' not in line)

def is_line_from_import(line):
    return bool(line and line.startswith(from_tag) and ' import ' in line)

def is_from_group_start(line):
    return bool(line and line.startswith(from_tag) and 'import' in line and '(' in line and not line.rstrip().endswith(')'))

def is_from_group_end(line):
    return bool(line and ')' in line)

def clean_imports(imports):
    if isinstance(imports, str):
        imports = imports.split(',')
    return [eatElse(imp.strip()) for imp in imports if imp.strip()]

# ============================================================
# Combine lone import statements
# ============================================================
def combine_lone_imports(text=None, file_path=None):
    text = text or ''
    if file_path and os.path.isfile(file_path):
        text += read_from_file(file_path)
    lines = text.split('\n')

    cleaned_import_list = []
    nu_lines = []
    j = None

    for i, line in enumerate(lines):
        if is_line_import(line):
            if j is None:
                nu_lines.append(import_tag)
                j = i
            cleaned_import_list += clean_imports(line.split(import_tag)[1])
        else:
            nu_lines.append(line)

    if j is None:
        return '\n'.join(nu_lines)
    cleaned_import_list = sorted(set(cleaned_import_list))
    nu_lines[j] += ', '.join(cleaned_import_list)
    return '\n'.join(nu_lines)

# ============================================================
# Merge repeated 'from pkg import ...' (1-line only)
# Preserve multi-line grouped imports
# ============================================================
def merge_from_import_groups(text=None, file_path=None):
    if file_path and os.path.isfile(file_path):
        text = read_from_file(file_path)
    text = text or ''
    lines = text.split('\n')

    pkg_to_imports: Dict[str, Set[str]] = {}
    pkg_to_line_index: Dict[str, int] = {}
    nu_lines: List[str] = []

    in_group = False
    for i, line in enumerate(lines):
        stripped = line.strip()

        # preserve multi-line grouped blocks intact
        if in_group:
            nu_lines.append(line)
            if is_from_group_end(line):
                in_group = False
            continue

        if is_from_group_start(line):
            in_group = True
            nu_lines.append(line)
            continue

        if is_line_from_import(line):
            try:
                pkg_part, imps_part = line.split(' import ', 1)
                pkg_name = pkg_part.replace('from ', '').strip()
                imps = clean_imports(imps_part)
            except Exception:
                nu_lines.append(line)
                continue

            if pkg_name not in pkg_to_imports:
                pkg_to_imports[pkg_name] = set(imps)
                pkg_to_line_index[pkg_name] = len(nu_lines)
                nu_lines.append(line)
            else:
                pkg_to_imports[pkg_name].update(imps)
        else:
            nu_lines.append(line)

    # Rewrite first occurrences
    for pkg, idx in pkg_to_line_index.items():
        all_imps = sorted(pkg_to_imports[pkg])
        nu_lines[idx] = f"from {pkg} import {', '.join(all_imps)}"

    return '\n'.join(nu_lines)

# ============================================================
# Pipeline
# ============================================================
def clean_imports_pipeline(path: str):
    raw = read_from_file(path)
    step1 = combine_lone_imports(text=raw)
    step2 = merge_from_import_groups(text=step1)
    return step2

# ============================================================
# Standalone Run
# ============================================================
if __name__ == "__main__":
    abs_path = "/home/flerb/Documents/pythonTools/modules/src/modules/abstract_utilities/src/abstract_utilities/file_utils/imports/imports.py"
    cleaned = clean_imports_pipeline(abs_path)
    print(cleaned)



import os
from collections import defaultdict
from ...imports import hashlib
from ...json_utils import safe_dump_to_json
from ...list_utils import make_list
from ...file_utils import get_files_and_dirs



from ...list_utils import make_list
from ...type_utils import get_media_exts, is_media_type, MIME_TYPES, is_str,if_not_bool_default
from ...ssh_utils import *
from ...env_utils import *
from ...read_write_utils import read_from_file,write_to_file
from ...log_utils import get_logFile
from ...class_utils import if_none_default,if_none_change,get_caller,get_initial_caller_dir, get_caller_path, get_caller_dir,SingletonMeta,run_pruned_func
from ...string_utils import eatAll
